/*  util.c --- various utility functions that don't end up in the library

 Copyright (C) 2009 by Walter C. Pelissero

 Author: Walter C. Pelissero <walter@pelissero.de>
 Project: igorplug

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2, or (at
your option) any later version.
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.
You should have received a copy of the GNU General Public License
along with this program; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.*/

#include <stdio.h>
#include <strings.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/time.h>

#include "igorplug.h"
#include "util.h"

bool debug = 0;

void
print_packet (const IRbuffer buf, FILE *fp)
{
  int i;

  fprintf(fp, "%3d ", IGORPLUG_SERIAL(buf));
  fprintf(fp, "%3d ", IGORPLUG_OFFSET(buf));
  for (i = 0; i < IGORPLUG_DATA_LENGTH(buf); ++i)
    fprintf(fp, "%02X", IGORPLUG_DATA(buf)[i]);
  fprintf(fp, "\n");
}

static int
hexchar_to_int (char c)
{
  const static char digits[] = "0123456789ABCDEF";
  char *p = strchr(digits, toupper(c));
  if (p)
    return p - digits;
  return -1;
}

static int
parse_IR_code (char **string, IRbyte *code, unsigned max_length)
{
  int i;

  for (i = 0; i < max_length; ++i)
    {
      int x, y;

      x = hexchar_to_int(**string);
      if (x < 0)
	break;
      ++(*string);
      y = hexchar_to_int(**string);
      if (y < 0)
	break;
      ++(*string);
      code[i] = x << 4 | y;
    }
  return i;
}

static void
print_byte_sequence (IRbyte *seq, unsigned length, FILE *fp)
{
  int i;
  for (i = 0; i < length; ++i)
    fprintf(fp, "%02X", seq[i]);
}


static void
print_action_entry (const action_entry *entry, FILE *fp)
{
  fprintf(fp, "action_entry 0x%X: code=", (unsigned int) entry);
  print_byte_sequence(entry->code, entry->code_length, fp);
  fprintf(fp, ", min_pause=%d\n", entry->min_pause);
  fprintf(fp, ", action=0x%X\n", (unsigned int) entry->action);
}

char *
trim_whitespace (char *string)
{
  while (*string && isspace(*string))
    ++string;
  if (*string)
    {
      char *p = string, *end = string;

      while (*p)
	{
	  if (!isspace(*p))
	    end = p;
	  ++p;
	}
      *(end + 1) = '\0';
    }
  return string;
}

action_entry *
read_config_file (const char *pathname, action_parser parser)
{
  FILE *fp;
  action_entry first, *ap = &first;
  char line[1024];

  if (!(fp = fopen(pathname, "r")))
    {
      fprintf(stderr, "can't open \"%s\" (errno=%d)\n", pathname, errno);
      return 0;
    }
  if (!parser)
    parser = (action_parser) strdup;
  first.next = 0;
  while (fgets(line, sizeof(line), fp))
    {
      size_t l;

      line[sizeof(line) - 1] ='\0';
      dprint("conf: %s", line);
      l = strspn(line, " \t");
      if (l >= 0 && !(line[l] == '#'))
	{
	  char *start = line + l;
	  IRbyte code[1024];
	  int length;
	  void *action;
	  unsigned min_pause = 0;

	  length = parse_IR_code(&start, code, sizeof(code));
	  if (*start == ',')
	    min_pause = strtol(start + 1, &start, 10);
	  action = parser(trim_whitespace(start));
	  if (action && length > 0)
	    {
	      action_entry *entry = malloc(sizeof(action_entry));

	      entry->code_length = length;
	      entry->code = malloc(length);
	      bcopy(code, entry->code, length);
	      entry->min_pause = min_pause;
	      entry->action = action;
	      if (debug)
		print_action_entry(entry, stderr);
	      ap->next = entry;
	      ap = entry;
	    }
	}
    }
  ap->next = 0;
  return first.next;
}

action_entry *
find_action (IRbuffer buffer, action_entry *actions, unsigned tolerance)
{
  action_entry *p = actions;

  while (p)
    {
      if (IGORPLUG_DATA_LENGTH(buffer) == p->code_length &&
	  compare_codes(p->code, IGORPLUG_DATA(buffer), p->code_length, tolerance))
	return p;
      p = p->next;
    }
  return 0;
}

void
dprint (const char *fmt, ...)
{
  if (debug)
    {
      va_list ap;
      va_start(ap, fmt);
      vfprintf(stderr, fmt, ap);
      va_end(ap);
    }
}

static unsigned long
timeval2long (struct timeval *tp)
{
  return tp->tv_sec * 1000 + tp->tv_usec / 1000;
}

static unsigned long
get_epoch ()
{
  struct timeval tp;
  gettimeofday(&tp, 0);
  return timeval2long(&tp);
}

int
main_loop (int fd, action_executor execute, action_entry *actions, unsigned tolerance, void *data)
{
  unsigned char *buf = igorplug_alloc_buffer();
  action_entry *previous_entry = 0;
  unsigned long previous_epoch;

  while (1)
    {
      action_entry *entry;

      if (igorplug_read_buffer(fd, buf) < 0)
	{
	  fprintf(stderr, "igorplug_read_buffer failed (errno=%d)\n", errno);
	  exit(2);
	}
      if (debug)
	print_packet(buf, stdout);
      entry = find_action(buf, actions, tolerance);
      dprint("entry=0x%X\n", (unsigned) entry);
      if (entry)
	{
	  unsigned long epoch = get_epoch();

	  if (debug)
	    print_action_entry(entry, stderr);
	  if (!previous_entry ||
	      entry != previous_entry ||
	      entry->min_pause <= (epoch - previous_epoch))
	    execute(entry->action, data);
	  previous_epoch = epoch;
	}
      previous_entry = entry;
    }
}
